import numpy as np

from pybullet_industrial.interpolation import linear_interpolation, circular_interpolation


def build_box_path(center_position: np.array, dimensions: np.array,
                   radius: float, orientation: np.array, samples: int):
    """Build a box shaped path with rounded corners

    Args:
        center_position (np.array): The 3D center position of the box
        dimensions (np.array): the length and width of the box
        radius (float): the radius of the corners
        orientation (np.array): A quaternion describing the orientation of the box
        samples (int): The number of points in the box path

    Returns:
        ToolPath: The resulting Toolpath
    """
    corner_point_00 = np.array(
        [-0.5*dimensions[0], -0.5*dimensions[1]+radius, 0])
    corner_point_01 = np.array(
        [-0.5*dimensions[0], 0.5*dimensions[1]-radius, 0])
    corner_point_10 = np.array(
        [-0.5*dimensions[0]+radius, 0.5*dimensions[1], 0])
    corner_point_11 = np.array(
        [0.5*dimensions[0]-radius, 0.5*dimensions[1], 0])
    corner_point_20 = np.array(
        [0.5*dimensions[0], 0.5*dimensions[1]-radius, 0])
    corner_point_21 = np.array(
        [0.5*dimensions[0], -0.5*dimensions[1]+radius, 0])
    corner_point_30 = np.array(
        [0.5*dimensions[0]-radius, -0.5*dimensions[1], 0])
    corner_point_31 = np.array(
        [-0.5*dimensions[0]+radius, -0.5*dimensions[1], 0])

    linear_length = (np.linalg.norm(corner_point_00-corner_point_01) +
                     np.linalg.norm(corner_point_10-corner_point_11) +
                     np.linalg.norm(corner_point_20-corner_point_21) +
                     np.linalg.norm(corner_point_30-corner_point_31))

    circle_length = 2*np.pi*radius

    linear_samples = int(0.25*linear_length /
                         (linear_length+circle_length)*samples)
    circle_samples = int(0.25*circle_length /
                         (linear_length+circle_length)*samples)

    box_path = linear_interpolation(
        corner_point_00, corner_point_01, linear_samples)
    side_1 = linear_interpolation(
        corner_point_10, corner_point_11, linear_samples)
    side_2 = linear_interpolation(
        corner_point_20, corner_point_21, linear_samples)
    side_3 = linear_interpolation(
        corner_point_30, corner_point_31, linear_samples)

    corner_0 = circular_interpolation(
        corner_point_01, corner_point_10, radius, circle_samples)
    corner_1 = circular_interpolation(
        corner_point_11, corner_point_20, radius, circle_samples)
    corner_2 = circular_interpolation(
        corner_point_21, corner_point_30, radius, circle_samples)
    corner_3 = circular_interpolation(
        corner_point_31, corner_point_00, radius, circle_samples)

    box_path.append(corner_0)
    box_path.append(side_1)
    box_path.append(corner_1)
    box_path.append(side_2)
    box_path.append(corner_2)
    box_path.append(side_3)
    box_path.append(corner_3)

    box_path.rotate(orientation)
    box_path.translate(center_position)

    return box_path



from typing import Dict

import numpy as np
import pybullet as p


from pybullet_industrial.endeffector_tool import EndeffectorTool
from pybullet_industrial.robot_base import RobotBase


class MillingTool(EndeffectorTool):
    """A class implementing a milling tool.
    The tool is modeled as a set of teeth that remove material
    and apply forces on the tcp frame according to the kienzle model.

    Args:
        urdf_model (str): The path to the urdf model of the milling tool
        start_position (np.array): The position of the tool center point
        start_orientation (np.array): The orientation of the tool center point
        milling_properties (Dict): A dictionary containing the properties of the milling tool.
                                   Default values are:
                                   'diameter':0.05, 'height':0.01,
                                   'number of rays':10,'rotation speed':0.1, 'number of teeth':5
                                   'material_specific_force':2500,'chip_thickness_exponent':0.26
                                   The parameters kc11 and m_c are specific for the material.
        coupled_robot (RobotBase, optional): A pybullet_industrial.RobotBase object if
                                             the robot is coupled from the start.
                                             Defaults to None.
        tcp_frame (str, optional): The name of the urdf_link
                                   describing the tool center point.
                                   Defaults to None in which case the last link is used.
        connector_frame (str, optional): The name of the urdf_link
                                         at which a robot connects.
                                         Defaults to None in which case the base link is used


    """

    def __init__(self, urdf_model: str, start_position: np.array, start_orientation: np.array,
                 milling_properties: Dict, coupled_robot: RobotBase = None,
                 tcp_frame: str = None, connector_frame: str = None):

        super().__init__(urdf_model, start_position, start_orientation,
                         coupled_robot, tcp_frame, connector_frame)

        self.properties = {'diameter': 0.05,
                           'rotation speed': 1,
                           'number of teeth': 5,
                           'height': 0.1,
                           'number of rays': 10,
                           'material_specific_force': 2500,
                           'chip_thickness_exponent': 0.26}
        self.current_angle = 0
        if milling_properties is not None:
            self.change_properties(milling_properties)

    def get_cutting_state(self, ray_cast_result, tcp_frame=None):
        """A helpfer function calculating the cutting depth and speed of the tool.

        Args:
            ray_cast_result (list): The result of a batch ray cast as performed
                                      by the mill function
            tcp_frame (str, optional): The tool center point frame name. Defaults to None.

        Returns:
            float: the speed at which the cutting tool is moved into the material
            (np.array): an array of the depth of the cutting tool at each tooth
        """
        if tcp_frame is None:
            tcp_id = self._tcp_id
        else:
            tcp_id = self._convert_link_to_id(tcp_frame)

        link_state = p.getLinkState(self.urdf, tcp_id, computeLinkVelocity=1)
        cutting_speed = np.linalg.norm(link_state[7])

        ray_cast_per_teeth = [ray_cast_result[i:i+self.properties['number of rays']]
                              for i in range(0, len(ray_cast_result),
                                             self.properties['number of rays'])]

        # calculate the cutting depth based on the amount of rays that are hitting a body
        # if the cutting depth is zero, the tool is not cutting
        cutting_depth = [sum([ray_cast[0] != -1 for ray_cast in ray_cast_per_teeth[i]])
                         for i in range(self.properties['number of teeth'])]
        cutting_depth = np.array(
            cutting_depth)*self.properties['height']/self.properties['number of rays']

        return cutting_speed, cutting_depth

    def cast_rays(self, position: np.array, orientation: np.array, debug=False):
        """Randomly casts rays withn a given range from a specified position and orientation

        Args:
            position (np.array): start position of the raycast
            orientation (np.array): start orientation of the raycast
            debug (bool, optional): If true, the debug lines are drawn. Defaults to False.

        Returns:
            List: The result of a raycast
        """
        ray_start_pos = []
        ray_end_pos = []
        angle_between_teeth = 2*np.pi/self.properties['number of teeth']

        rot_matrix = p.getMatrixFromQuaternion(orientation)
        rot_matrix = np.array(rot_matrix).reshape(3, 3)

        for i in range(self.properties['number of teeth']):
            end_point = self.properties['diameter']*np.array(
                [np.sin(i*angle_between_teeth+self.current_angle),
                 np.cos(i*angle_between_teeth+self.current_angle),
                 0])
            for j in range(self.properties['number of rays']):
                height_adjustment = np.array(
                    [0, 0, j*self.properties['height']/self.properties['number of rays']])

                start_position = position+rot_matrix@height_adjustment
                end_position = position + \
                    rot_matrix@(end_point+height_adjustment)

                if debug:
                    p.addUserDebugLine(start_position, end_position,
                                       [1, 0, 0], 1, lifeTime=1)
                ray_start_pos.append(start_position)
                ray_end_pos.append(end_position)

        ray_cast_results = p.rayTestBatch(ray_start_pos, ray_end_pos)

        return ray_cast_results

    def calculate_process_force(self, ray_cast_results, tcp_frame=None):
        """Function calculating the process force of the milling tool.

        Args:
            ray_cast_results (list): The result of a batch ray cast
                                       as performed by the mill function
            tcp_frame (str): The name of the tool center point frame. Defaults to None.

        Returns:
            np.array: a 3 dimensional array containing the process force in the x,y,z direction
        """
        cutting_speed, cutting_depth = self.get_cutting_state(
            ray_cast_results, tcp_frame)
        teeth_angles = [i * 2*np.pi/self.properties['number of teeth'] +
                        self.current_angle for i in range(self.properties['number of teeth'])]
        cutting_force = self.force_model(cutting_speed,
                                         cutting_depth,
                                         teeth_angles)
        return cutting_force

    def mill(self, tcp_frame=None, debug=False):
        """Function that performs a milling operation.

        Args:
            tcp_frame (str, optional): The name of the tool center point frame. Defaults to None.
            debug (bool, optional): If true, the debug lines are drawn. Defaults to False.

        Returns:
            list: A list containing the ids of the bodies that were removed
        """

        position, orientation = self.get_tool_pose(tcp_frame)
        ray_cast_results = self.cast_rays(position, orientation, debug)

        cutting_force = self.calculate_process_force(
            ray_cast_results, tcp_frame)
        self.apply_tcp_force(cutting_force, tcp_frame)

        self.current_angle = self.current_angle + \
            self.properties['rotation speed']

        removed_objects = []
        for ray_intersection in ray_cast_results:
            if ray_intersection[0] != -1:
                p.removeBody(ray_intersection[0])
                removed_objects.append(ray_intersection[0])
        return removed_objects

    def force_model(self, cutting_speed, cutting_depth, teeth_angles):
        """A force model that is used to calculate the force that is applied to the tool.
        Args:
            cutting_speed (float): the speed at which the cutting tool is moved into the material
            cutting_depth (np.array): an array of the depth of the cutting tool at each tooth
            teeth_angles (list): the angles of the teeth on the tool
        Returns:
            np.array: an array of the force that is applied to the cutting tool at the tcp
        """
        h = cutting_speed / \
            (self.properties['rotation speed'] *
             self.properties['number of teeth'])
        k_c = self.properties['material_specific_force'] / \
            (h ** self.properties['chip_thickness_exponent'])

        force = np.zeros(3)

        for index, depth in enumerate(cutting_depth):
            cutting_force = k_c * depth * h
            force -= np.array([cutting_force*np.sin(teeth_angles[index]),
                               cutting_force*np.cos(teeth_angles[index]), 0])

        return force

    def change_properties(self, new_properties: Dict):
        """Allows retroactive changes to the ray casting properties.

        Args:
            new_properties (Dict): A dictionary containing values and keys that
                                   should be changed
        Raises:
            KeyError: If a key is not a valid property
        """
        for key in new_properties:
            if not key in self.properties:
                raise KeyError("The specified property keys are not valid" +
                               " Valid keys are: "+str(self.properties.keys()))
            self.properties[key] = new_properties[key]



from typing import Dict

import numpy as np
import pybullet as p


class RobotBase:
    """A Base class encapsulating a URDF based industrial robot manipulator

    Args:
        urdf_model (str): A valid path to a urdf file
        start_position (np.array): The start position of the robot base
        start_orientation (np.array): A quaternion describing the start orientation
                                      of the robot base
        default_endeffector (str, optional): The default endeffector used
                                             when controlling the robots position
    """

    def __init__(self, urdf_model: str, start_position: np.array, start_orientation: np.array,
                 default_endeffector: str = None):

        urdf_flags = p.URDF_USE_SELF_COLLISION_EXCLUDE_ALL_PARENTS
        self.urdf = p.loadURDF(urdf_model,
                               start_position, start_orientation,
                               flags=urdf_flags,
                               useFixedBase=True)

        self.number_of_joints = p.getNumJoints(self.urdf)
        self._joint_state_shape = self.get_joint_state()
        self._joint_name_to_index = {}
        self._link_name_to_index = {}
        kinematic_solver_map = []
        self._lower_joint_limit = np.zeros(self.number_of_joints)
        self._upper_joint_limit = np.zeros(self.number_of_joints)

        for joint_number in range(self.number_of_joints):
            joint_info = p.getJointInfo(self.urdf, joint_number)
            link_name = joint_info[12].decode("utf-8")
            self._link_name_to_index[link_name] = joint_number

            if joint_info[2] != 4:  # checks if the joint is not fixed
                joint_name = joint_info[1].decode("utf-8")
                self._joint_name_to_index[joint_name] = joint_number

                kinematic_solver_map.append(joint_number)

                lower_limit = joint_info[8]
                upper_limit = joint_info[9]
                if upper_limit < lower_limit:
                    lower_limit = -np.inf
                    upper_limit = np.inf
                self._lower_joint_limit[joint_number] = lower_limit
                self._upper_joint_limit[joint_number] = upper_limit

        self._kinematic_solver_map = np.array(kinematic_solver_map)

        if default_endeffector is None:
            last_link = max(self._link_name_to_index)
            self._default_endeffector_id = self._link_name_to_index[last_link]
        else:
            self._default_endeffector_id = self._convert_endeffector(
                default_endeffector)

        self.max_joint_force = 1000*np.ones(self.number_of_joints)
        for joint_number in range(self.number_of_joints):
            p.resetJointState(self.urdf, joint_number, targetValue=0)


    def get_joint_state(self):
        """Returns the position of each joint as a dictionary keyed with their name

        Returns:
             Dict[str,Dict[str,float]]: The state of all joinst

        """
        joint_state = {}
        for joint_number in range(self.number_of_joints):
            joint_info = p.getJointInfo(self.urdf, joint_number)
            if joint_info[2] != 4:  # checks if the joint is not fixed
                # convert byte string to string
                joint_name = joint_info[1].decode("utf-8")
                joint_state_list = p.getJointState(self.urdf, joint_number)

                single_joint_state = {'position': joint_state_list[0],
                                      'velocity': joint_state_list[1],
                                      'reaction force': joint_state_list[2],
                                      'torque': joint_state_list[3]}
                joint_state[joint_name] = single_joint_state
        return joint_state

    def set_joint_position(self, target: Dict[str,  float], ignore_limits=False):
        """Sets the target position for a number of joints.
           The maximum force of each joint is set according to the max_joint_force class attribute.

        Args:
            target (Dict[str,  float]): A dictionary containing the joint states to be set

        Raises:
            KeyError: If the specified joint state is not part of the Robot
        """
        if not all(key in self._joint_state_shape for key in target):
            raise KeyError('One or more joints are not part of the robot. ' +
                           'correct keys are: '+str(self._joint_state_shape.keys()))

        for joint, joint_position in target.items():
            joint_number = self._joint_name_to_index[joint]

            if ignore_limits is False:
                lower_joint_limit = self._lower_joint_limit[joint_number]
                upper_joint_limit = self._upper_joint_limit[joint_number]
                if joint_position > upper_joint_limit or joint_position < lower_joint_limit:
                    raise ValueError('The joint position '+str(joint_position) +
                                     ' is out of limit for joint '+joint+'. Its limits are:\n' +
                                     str(lower_joint_limit)+' and '+str(upper_joint_limit))

            p.setJointMotorControl2(self.urdf, joint_number, p.POSITION_CONTROL,
                                    force=self.max_joint_force[joint_number],
                                    targetPosition=joint_position)

    def get_endeffector_pose(self, endeffector_name: str = None):
        """Returns the position of the endeffector in world coordinates

        Args:
            endeffector (str, optional): The name of a different endeffector link

        Returns:
            np.array: The position of the endeffector
            np.array: The orientation of the endeffector as a quaternion
        """
        if endeffector_name is None:
            endeffector_id = self._default_endeffector_id
        else:
            endeffector_id = self._convert_endeffector(endeffector_name)

        link_state = p.getLinkState(self.urdf, endeffector_id)

        position = np.array(link_state[0])
        orientation = np.array(link_state[1])
        return position, orientation

    def set_endeffector_pose(self, target_position: np.array, target_orientation: np.array = None,
                             endeffector_name: str = None):
        """Sets the pose of a robots endeffector

        Args:
            target_position (np.array): The desired 3D position
            target_orientation (np.array, optional): The desired orientation as a quaternion.
                                                     Defaults to None.
            endeffector_name (str, optional): The name of a different endeffector.
                                              Defaults to None.
        """
        if endeffector_name is None:
            endeffector_id = self._default_endeffector_id
        else:
            endeffector_id = self._convert_endeffector(endeffector_name)

        if target_orientation is None:
            joint_poses = p.calculateInverseKinematics(self.urdf,
                                                       endeffector_id,
                                                       target_position,
                                                       lowerLimits=self._lower_joint_limit,
                                                       upperLimits=self._upper_joint_limit)
        else:
            joint_poses = p.calculateInverseKinematics(self.urdf,
                                                       endeffector_id,
                                                       target_position,
                                                       targetOrientation=target_orientation,
                                                       lowerLimits=self._lower_joint_limit,
                                                       upperLimits=self._upper_joint_limit)

        for index, joint_position in enumerate(joint_poses):
            joint_number = self._kinematic_solver_map[index]
            p.setJointMotorControl2(self.urdf, joint_number, p.POSITION_CONTROL,
                                    force=self.max_joint_force[joint_number],
                                    targetPosition=joint_position)

    def reset_robot(self, start_position: np.array, start_orientation: np.array,
                    joint_values: list = None):
        """resets the robots joints to 0 and the base to a specified position and orientation

        Args:
            start_position (np.array): a 3 dimensional position
            start_orientation (np.array): a 4 dimensional quaternion representing
                                          the desired orientation
            joint_values (list): Allows to reset the joint state of the robot given
                                 a list of positions.
                                 Defaults to None in which case the joints remain in their current
                                 configuration.
        """
        self.set_world_state(start_position, start_orientation)

        if joint_values is None:
            joint_values = np.zeros(self.number_of_joints)
        for joint in range(self.number_of_joints):
            p.resetJointState(self.urdf, joint,
                              targetValue=joint_values[joint])

    def set_world_state(self, start_position: np.array, start_orientation: np.array):
        """Resets the robots base to a specified position and orientation

        Args:
            start_position (np.array): a 3 dimensional position
            start_orientation (np.array): a 4 dimensional quaternion representing
                                          the desired orientation
        """
        p.resetBasePositionAndOrientation(
            self.urdf, start_position, start_orientation)

    def get_world_state(self):
        """Returns the position and orientation of the robot relative to the world

        Returns:
            list: the 3 dimensional position vector of the robot base
            list: a 4 dimensional quaternion representing the orientation of the robot base
        """
        return p.getBasePositionAndOrientation(self.urdf)

    def _convert_endeffector(self, endeffector: str):
        """Internal Function which converts an endeffector name to an id

        Args:
            endeffector (str): The name of the endeffector link

        Raises:
            TypeError: if the name is not a string.
            ValueError: if the endeffector name is not valid

        Returns:
            int: The corresponding link index to the endeffector id.
        """
        if not isinstance(endeffector, str):
            raise TypeError(
                "The Endeffector must be a String describing a URDF link")
        if not endeffector in self._link_name_to_index:
            raise ValueError("Invalid Endeffecot name! valid names are: " +
                             str(self._link_name_to_index.keys()))

        return self._link_name_to_index[endeffector]



from typing import Dict

import numpy as np
import pybullet as p


class Particle():
    """A template class for material particles extruded by a extruder endeffector tool.

    All materials should inherit from this class and implement its methods.

    Args:
        ray_cast_result (list): The result of a pybullet ray_cast
                                as performed by the Extruder class.
                                It is made up of: [objectUniqueId,
                                                    linkIndex,
                                                    hit fraction,
                                                    hit position,
                                                    hit normal]
        material_properties (Dict): A dictionary containing the properties of the material
    """

    def __init__(self, ray_cast_result: list, material_properties: Dict):

        self.properties = {}

    def get_position(self):
        """Returns the position of a particle in the world frame
        """
        pass

    def remove(self):
        """Function to actively remove the particle from the simulation.
           This function is deliberatly kept seperate from the __del__ method to prevent having
           to manually save particles if there is no intention of removing them.
        """
        pass

    def set_material_properties(self, new_properties: Dict):
        """Checks if the matieral properties contain the proper keys


        Args:
            new_properties (Dict): A dictionary containing the material properties

        Raises:
            KeyError: If a key is not a valid extruder property
        """
        for key in new_properties:
            if not key in self.properties:
                raise KeyError("The specified property keys are not valid" +
                               " Valid keys are: "+str(self.properties.keys()))
            self.properties[key] = new_properties[key]


class Plastic(Particle):
    """A class for simple particles which can be used for additive manufacturing.
        The particles are infinitely rigid and stick to each other.

    Args:
        ray_cast_result (list): The result of a pybullet ray_cast
                                as performed by the Extruder class.
                                It is made up of: [objectUniqueId,
                                                    linkIndex,
                                                    hit fraction,
                                                    hit position,
                                                    hit normal]
        material_properties (Dict): A dictionary containing the properties of the material.
                                    The default properties for Plastic are:
                                    'particle size': 0.3, 'color': [1, 0, 0, 1]
    """

    def __init__(self, ray_cast_result: list,  material_properties: Dict):

        self.properties = {'particle size': 0.3, 'color': [1, 0, 0, 1]}
        self.set_material_properties(material_properties)
        particle_size = self.properties['particle size']
        color = self.properties['color']

        visual_shape_id = p.createVisualShape(
            shapeType=p.GEOM_SPHERE, rgbaColor=color, radius=particle_size)
        collision_shape_id = p.createCollisionShape(
            shapeType=p.GEOM_SPHERE, radius=particle_size)

        self.particle_id = p.createMultiBody(baseMass=0,
                                             baseCollisionShapeIndex=collision_shape_id,
                                             baseVisualShapeIndex=visual_shape_id,
                                             basePosition=ray_cast_result[3])

    def get_position(self):
        """Returns the position of a particle in the world frame

        Returns:
            [float,float,float]: The three dimensional position of the particle
                                 in the world coordinate system
        """
        position, _ = p.getBasePositionAndOrientation(self.particle_id)
        return position

    def remove(self):
        p.removeBody(self.particle_id)


class Paint(Particle):
    """A class for simple Paint particles which stick to objects and move with them.
        The Paint particles are purely visible and have neither mass nor a collision mesh

    Args:
        ray_cast_result (list): The result of a pybullet ray_cast as performed by the extruder
        material_properties (Dict): A dictionary containing the properties of the material.
                                    The default properties for Paint are:
                                    'particle size': 0.3, 'color': [1, 0, 0, 1]
    """

    def __init__(self, ray_cast_result: list, material_properties: Dict):

        self.properties = {'particle size': 0.3, 'color': [1, 0, 0, 1]}
        self.set_material_properties(material_properties)
        particle_size = self.properties['particle size']
        color = self.properties['color']

        self.particle_ids = []
        self.target_id = ray_cast_result[0]
        self.target_link_id = ray_cast_result[1]
        self.hit_position = ray_cast_result[3]
        self.initial_target_position = self.get_target_pose(
            self.target_id, self.target_link_id)[0]

        if self.target_id != -1:
            target_position, target_orientation = self.get_target_pose(
                self.target_id, self.target_link_id)

            rot_matrix = p.getMatrixFromQuaternion(target_orientation)
            rot_matrix = np.array(rot_matrix).reshape(3, 3)

            self.adj_target_position = np.linalg.inv(
                rot_matrix)@(np.array(ray_cast_result[3])-target_position)

            steps = 3
            width = particle_size*500
            theta2 = np.linspace(-np.pi,  0, steps)
            phi2 = np.linspace(0,  5 * 2*np.pi, steps)

            x_coord = particle_size * \
                np.sin(theta2) * np.cos(phi2) + self.adj_target_position[0]
            y_coord = particle_size * \
                np.sin(theta2) * np.sin(phi2) + self.adj_target_position[1]
            z_coord = particle_size * \
                np.cos(theta2) + self.adj_target_position[2]

            path = np.array([x_coord, y_coord, z_coord])
            path_steps = len(path[0])
            for i in range(1, path_steps):
                current_point = path[:, i]
                previous_point = path[:, i-1]
                self.particle_ids.append(p.addUserDebugLine(current_point, previous_point,
                                                            lineColorRGB=color[:3],
                                                            lineWidth=width,
                                                            lifeTime=0,
                                                            parentObjectUniqueId=self.target_id,
                                                            parentLinkIndex=self.target_link_id))

    @staticmethod
    def get_target_pose(target_id: int, target_link_id: int):
        """Returns the pose of a target objects link.

        This function is used to calculate the relative position
        of the paint particle to the object it sticks to.

        Args:
            target_id (int): The unique id of the target object
            target_link_id (int): The link id of the target object

        Returns:
            np.array, np.array: The position and orientation of the target object
        """
        if target_link_id == -1:
            target_position, target_orientation = p.getBasePositionAndOrientation(
                target_id)
        else:
            target_link_state = p.getLinkState(target_id, target_link_id)
            target_position = np.array(target_link_state[0])
            target_orientation = np.array(target_link_state[1])
        return target_position, target_orientation

    def get_position(self):
        """Returns the position of a particle in the world frame

        Returns:
            [float,float,float]: The three dimensional position of the particle
                                 in the world coordinate system
        """
        hit_position = self.hit_position
        initial_target_position = self.initial_target_position
        diff_vector = np.array(hit_position)-np.array(initial_target_position)

        target_position, target_orientation = self.get_target_pose(
            self.target_id, self.target_link_id)

        rot_matrix = p.getMatrixFromQuaternion(target_orientation)
        rot_matrix = np.array(rot_matrix).reshape(3, 3)

        coordinates = target_position+rot_matrix @ (diff_vector)

        return coordinates

    def remove(self):
        """Function to actively remove the particle from the simulation.
           This function is deliberatly kept seperate from the __del__ method to prevent having
           to manually save particles if there is no intention of removing them.
        """
        [p.removeUserDebugItem(id) for id in self.particle_ids]


class MetalVoxel(Particle):
    """A simple voxel class for cutting and milling simulations

    Args:
        ray_cast_result (list): The result of a pybullet ray_cast
                                as performed by the Extruder class.
        material_properties (Dict): A dictionary containing the properties of the material.
                                    The default properties for a Metal Voxel are:
                                    'particle size': 0.3, 'color': [1, 0, 0, 1]
    """

    def __init__(self, ray_cast_result: list,  material_properties: Dict):

        self.properties = {'particle size': 0.3, 'color': [1, 0, 0, 1]}
        self.set_material_properties(material_properties)
        particle_size = self.properties['particle size']
        color = self.properties['color']

        half_extents = particle_size*0.5
        visual_shape_id = p.createVisualShape(shapeType=p.GEOM_BOX,
                                              rgbaColor=color,
                                              halfExtents=[half_extents,
                                                           half_extents,
                                                           half_extents])
        collision_shape_id = p.createCollisionShape(shapeType=p.GEOM_BOX,
                                                    halfExtents=[half_extents,
                                                                 half_extents,
                                                                 half_extents])

        self.particle_id = p.createMultiBody(baseMass=0,
                                             baseCollisionShapeIndex=collision_shape_id,
                                             baseVisualShapeIndex=visual_shape_id,
                                             basePosition=ray_cast_result[3])

    def get_position(self):
        """Returns the position of a particle in the world frame

        Returns:
            [float,float,float]: The three dimensional position of the particle
                                 in the world coordinate system
        """
        position, _ = p.getBasePositionAndOrientation(self.particle_id)
        return position

    def remove(self):
        p.removeBody(self.particle_id)


def spawn_material_block(base_position: list, dimensions: list,
                         material: Particle, material_properties: Dict):
    """Spawns a block of a give material.

    Args:
        base_position ([float,float,float]): The position of the lower left base corner of the block
        dimensions ([float,float,float]): The dimensions of the block in [width,breath,height]
        material (Particle): A particle that should be spawned
        material_properties (Dict): A dictionary containing the properties of the material.
                                    It needs to contain a key 'particle size'.

    Returns:
        list[Particle]: A list of the spawned particles
    """
    if 'particle size' not in material_properties.keys():
        raise KeyError(
            "The material properties must contain the key 'particle size'!")
    particle_size = material_properties['particle size']
    half_extents = particle_size*0.5

    batch_positions = []

    for x in range(int(dimensions[0]/particle_size)):
        for y in range(int(dimensions[1]/particle_size)):
            for z in range(int(dimensions[2]/particle_size)):
                batch_positions.append(
                    [x * particle_size+base_position[0]+half_extents,
                     y * particle_size+base_position[1]+half_extents,
                     z * particle_size+base_position[2]+half_extents])

    objects = []
    for positions in batch_positions:
        particle = material([0, 0, 0, positions], material_properties)
        objects.append(particle)

    return objects



from typing import Dict

import numpy as np

from pybullet_industrial.material import Plastic
from pybullet_industrial.raycaster import RayCaster
from pybullet_industrial import RobotBase


class Extruder(RayCaster):
    """Special Endeffector Tool which can extrude material from its tcp.

    Args:
        urdf_model (str): A valid path to a urdf file describint the tool geometry
        start_position (np.array): the position at which the tool should be spawned
        start_orientation (np.array): the orientation at which the tool should be spawned
        extruder_properties(Dict): A dictionary containing the properties of the extrusion head.
                                   During initialization only 'material' has to be set.
                                   Default Values are:
                                   'opening angle':0,'number of rays':1,
                                   'material properties': {'particle size':0.03,
                                                           'color' : [1, 0, 0, 1]},
                                   'maximum distance':1,'material':Particle,
        coupled_robot (RobotBase, optional): A pybullet_industrial.RobotBase object if
                                             the robot is coupled from the start.
                                             Defaults to None.
        tcp_frame (str, optional): The name of the urdf_link
                                   describing the tool center point.
                                   Defaults to None in which case the last link is used.
        connector_frame (str, optional): The name of the urdf_link
                                         at which a robot connects.
                                         Defaults to None in which case the base link is used.

    Raises:
        ValueError: If no material is provided during initialization.
    """

    def __init__(self, urdf_model: str, start_position: np.array, start_orientation: np.array,
                 extruder_properties: Dict, coupled_robot: RobotBase = None,
                 tcp_frame: str = None, connector_frame: str = None):

        super().__init__(urdf_model, start_position, start_orientation,
                         coupled_robot, tcp_frame, connector_frame)

        self.properties['material'] = Plastic
        self.properties['material properties'] = {
            'particle size': 0.03, 'color': [1, 0, 0, 1]}

        self.change_properties(extruder_properties)

    def calculate_process_force(self, ray_cast_results, tcp_frame=None):
        """Function calculating the process force of the milling tool.

        Args:
            ray_cast_results (list): The result of a batch ray cast
                                       as performed by the remove function
            tcp_frame (str): The name of the tool center point frame. Defaults to None.

        Returns:
            np.array: a 3 dimensional array containing the process force in the x,y,z direction
        """

        return np.zeros(3)

    def extrude(self, tcp_frame: str = None):
        """Extrudes material from the specified tcp_frame.

        Args:
            tcp_frame (str, optional): the name of the link from which to extrude the material.
                                       Defaults to None in which case the default tcp is used
        """
        position, orientation = self.get_tool_pose(tcp_frame)
        ray_cast_results = self.cast_rays(position, orientation)

        process_force = self.calculate_process_force(
            ray_cast_results, tcp_frame)
        self.apply_tcp_force(process_force, tcp_frame)

        particle_list = []
        for ray_intersection in ray_cast_results:
            if ray_intersection[0] != -1:
                particle = self.properties['material'](ray_intersection,
                                                       self.properties['material properties'])
                particle_list.append(particle)
        return particle_list



from pybullet_industrial.robot_base import *
from pybullet_industrial.utility import *
from pybullet_industrial.endeffector_tool import *
from pybullet_industrial.sensors import *
from pybullet_industrial.gripper import *
from pybullet_industrial.extruder import *
from pybullet_industrial.material import *
from pybullet_industrial.raycaster import *
from pybullet_industrial.remover import *
from pybullet_industrial.toolpath import *
from pybullet_industrial.interpolation import circular_interpolation, linear_interpolation, spline_interpolation
from pybullet_industrial.path_builders import *
from pybullet_industrial.milling_tool import *



import numpy as np
import scipy.interpolate as sci

from pybullet_industrial.toolpath import ToolPath


def build_circular_path(center: np.array, radius: float,
                        min_angle: float, max_angle: float, step_num: int, clockwise: bool = True):
    """Function which builds a circular path

    Args:
        center (np.array): the center of the circle
        radius (float): the radius of the circle
        min_angle (float): minimum angle of the circle path
        max_angle (float): maximum angle of the circle path
        step_num (int): the number of steps between min_angle and max_angle
        clockwise (bool): boolean value indicating if the interpolation is performed clockwise
                          or anticlockwise

    Returns:
        np.array: array of 2 dimensional path points
    """

    circular_path = np.zeros((2, step_num))
    for j in range(step_num):
        if clockwise:
            path_angle = min_angle-j*(max_angle-min_angle)/step_num
        else:
            path_angle = min_angle+j*(max_angle-min_angle)/step_num
        new_position = center + radius * \
            np.array([np.cos(path_angle), np.sin(path_angle)])
        circular_path[:, j] = new_position
    return circular_path


def linear_interpolation(start_point: np.array, end_point: np.array, samples: int):
    """Performs a linear interpolation betwenn two points in 3D space

    Args:
        start_point (np.array): The start point of the interpolation
        end_point (np.array): The end point of the interpolation
        samples (int): The number of samples used to interpolate

    Returns:
        ToolPath: A ToolPath object of the interpolated path
    """
    final_path = np.linspace(start_point, end_point, num=samples)
    return ToolPath(final_path.transpose())


def planar_circular_interpolation(start_point: np.array, end_point: np.array,
                                  radius: float, samples: int, clockwise: bool = True):
    """Helper function which performs a circular interpolation in the x-y plane

    Args:
        start_point (np.array): The start point of the interpolation
        end_point (np.array): The end point of the interpolation
        radius (float): The radius of the circle
        samples (int): The number of samples used to interpolate
        clockwise (bool): boolean value indicating if the interpolation is performed clockwise
                            or anticlockwise

    Returns:
        np.array: The interpolated path
    """
    connecting_line = end_point-start_point
    distance_between_points = np.linalg.norm(connecting_line)
    if radius <= distance_between_points/2:
        raise ValueError("The radius needs to be at least " +
                         str(distance_between_points/2))

    center_distance_from_connecting_line = np.sqrt(
        radius**2-distance_between_points**2/4)

    if clockwise:
        orthogonal_vector = np.array(
            [connecting_line[1], -1*connecting_line[0]])
    else:
        orthogonal_vector = np.array(
            [-1*connecting_line[1], connecting_line[0]])

    circle_center = start_point+connecting_line/2+center_distance_from_connecting_line * \
        orthogonal_vector/np.linalg.norm(orthogonal_vector)

    angle_range = np.arccos(center_distance_from_connecting_line/radius)*2
    initial_angle = np.arctan2(
        start_point[1]-circle_center[1], start_point[0]-circle_center[0])

    planar_path = build_circular_path(
        circle_center, radius, initial_angle, initial_angle+angle_range, samples, clockwise)
    return planar_path


def circular_interpolation(start_point: np.array, end_point: np.array,
                           radius: float, samples: int, axis: int = 2, clockwise: bool = True):
    """AI is creating summary for circular_interpolation

    Args:
        start_point (np.array): The start point of the interpolation
        end_point (np.array): The end point of the interpolation
        radius (float): The radius of the circle used for the interpolation
        samples (int): The number of samples used to interpolate
        axis (int, optional): The axis around which the circle is interpolated.
                              Defaults to 2 which corresponds to the z-axis (0=x,1=y).
        clockwise (bool, optional): The direction of circular travel. Defaults to True.

    Returns:
        ToolPath: A ToolPath object of the interpolated path
    """

    all_axis = [0, 1, 2]
    all_axis.remove(axis)
    planar_start_point = np.array(
        [start_point[all_axis[0]], start_point[all_axis[1]]])
    planar_end_point = np.array(
        [end_point[all_axis[0]], end_point[all_axis[1]]])

    planar_path = planar_circular_interpolation(
        planar_start_point, planar_end_point, radius, samples, clockwise)

    path = np.zeros((3, samples))
    for i in range(2):
        path[all_axis[i]] = planar_path[i]
    path[axis] = np.linspace(start_point[axis], end_point[axis], samples)
    return ToolPath(path)


def spline_interpolation(points: np.array, samples: int):
    """Interpolates between a number of points in cartesian space.

    Args:
        points (np.array(3,n)): A 3 dimensional array whith each dimension containing
                                   subsequent positions.
        samples (int): The number of samples used to interpolate

    Returns:
        ToolPath: A ToolPath object of the interpolated path
    """
    s = np.linspace(0, 1, len(points[0]))

    path = np.zeros((3, samples))
    cs_x = sci.CubicSpline(s, points[0])
    cs_y = sci.CubicSpline(s, points[1])
    cs_z = sci.CubicSpline(s, points[2])

    cs_s = np.linspace(0, 1, samples)
    path[0] = cs_x(cs_s)
    path[1] = cs_y(cs_s)
    path[2] = cs_z(cs_s)

    return ToolPath(path)



import numpy as np
import pybullet as p

from pybullet_industrial import RobotBase


def draw_point(point: np.array, color: list = [0.0, 1.0, 0.0],
               length: float = 0.05, width: float = 2.0):
    """Draws a point in the worldspace as a cross of 3 lines

    Args:
        pnt (np.array): 3 dimensional point
        color (list, optional): RGB color. Defaults to [0.0,1.0,0.0].
        length (float, optional): The length of the lines.  Defaults to 0.5.
        width (float, optional): The width of the lines. Defaults to 2.0.
    """

    direction = np.array([1, 0, 0])
    p.addUserDebugLine(point + length*direction, point - length*direction,
                       lineColorRGB=color, lineWidth=width, lifeTime=0)

    direction = np.array([0, 1, 0])
    p.addUserDebugLine(point + length*direction, point - length*direction,
                       lineColorRGB=color, lineWidth=width, lifeTime=0)

    direction = np.array([0, 0, 1])
    p.addUserDebugLine(point + length*direction, point - length*direction,
                       lineColorRGB=color, lineWidth=width, lifeTime=0)


def draw_path(path: np.array, color: list = [0.0, 1.0, 0.0], width: float = 2.0):
    """Draws a path in the workspace

    Args:
        path (np.array(3,n)): Array containing the points in the path
        color (list, optional): RGB color. Defaults to [0.0,1.0,0.0].
        width (float, optional): The width of the lines. Defaults to 2.0.
    """
    path_steps = len(path[0])
    for i in range(1, path_steps):
        current_point = path[:, i]
        previous_point = path[:, i-1]
        p.addUserDebugLine(current_point, previous_point,
                           lineColorRGB=color, lineWidth=width, lifeTime=0)


def draw_coordinate_system(position: np.array, orientation: np.array, length: float = 0.1,
                           width: float = 2.0, life_time: float = 0, parent_id: int = -1,
                           parent_index: int = -1):
    """This function draws a coordinate system at a given position

    Args:
        position (np.array): The 3D position of the coordinate system
        orientation (np.array): The quaternion representing the orientation
                              of the coordinate system
        length (float, optional): The length of the lines.  Defaults to 0.5.
        width (float, optional): The width of the lines. Defaults to 2.0.
        life_time (float, optional): How long the coordinate system remains before despawning.
                                     Defaults to 0 in wich case it remains forever.
    """

    euler_angles = p.getEulerFromQuaternion(orientation)
    roll = euler_angles[0]
    pitch = -1*euler_angles[1]
    yaw = euler_angles[2]

    x_direction = np.array([np.cos(yaw)*np.cos(pitch),
                            np.sin(yaw)*np.cos(pitch),
                            np.sin(pitch)])
    y_direction = -1*np.array([-np.cos(yaw)*np.sin(pitch)*np.sin(roll)-np.sin(yaw)*np.cos(roll),
                               -np.sin(yaw)*np.sin(pitch) *
                               np.sin(roll)+np.cos(yaw)*np.cos(roll),
                               np.cos(pitch)*np.sin(roll)])
    z_direction = np.cross(x_direction, y_direction)

    p.addUserDebugLine(position,
                       position+length*x_direction,
                       lineColorRGB=[1, 0, 0],
                       lineWidth=width,
                       lifeTime=life_time,
                       parentObjectUniqueId=parent_id,
                       parentLinkIndex=parent_index)
    p.addUserDebugLine(position,
                       position+length*y_direction,
                       lineColorRGB=[0, 1, 0],
                       lineWidth=width,
                       lifeTime=life_time,
                       parentObjectUniqueId=parent_id,
                       parentLinkIndex=parent_index)
    p.addUserDebugLine(position,
                       position+length*z_direction,
                       lineColorRGB=[0, 0, 1],
                       lineWidth=width,
                       lifeTime=life_time,
                       parentObjectUniqueId=parent_id,
                       parentLinkIndex=parent_index)


def draw_robot_frames(robot: RobotBase, text_size: int = 1, length: float = 0.1,
                      width: float = 2.0, life_time: float = 0):
    """Visualizes the coordinate Frames of each link of a robot

    Args:
        robot (RobotBase): a Robot object
        text_size (int, optional): The size at which the text is rendered. Defaults to 1.
        length (float, optional): The length of the lines.  Defaults to 0.1.
        width (float, optional): The width of the lines. Defaults to 2.0.
        life_time (float, optional): How long the coordinate system remains before despawning.
                                     Defaults to 0 in wich case it remains forever.
    """
    for link_name, link_id in robot._link_name_to_index.items():
        orientation = p.getQuaternionFromEuler([0, 0, 0])
        draw_coordinate_system([0, 0, 0], orientation,
                               length, width, life_time, robot.urdf, link_id)
        p.addUserDebugText(link_name, [0, 0, 0], textSize=text_size, lifeTime=life_time,
                           parentObjectUniqueId=robot.urdf, parentLinkIndex=link_id)


def get_object_id_from_mouse():
    """Returns the object ID and Link ID of an object when clicking on it

    Returns:
        int: The Id of the object
        int: The Id of its link
    """
    mouse_events = p.getMouseEvents()
    for event in mouse_events:
        if ((event[0] == 2) and (event[3] == 0) and (event[4] & p.KEY_WAS_TRIGGERED)):
            mouse_x = event[1]
            mouse_y = event[2]

            debug_camera = p.getDebugVisualizerCamera()
            width = debug_camera[0]
            height = debug_camera[1]
            camera_forward = np.array(debug_camera[5])
            horizontal = np.array(debug_camera[6])
            vertical = np.array(debug_camera[7])
            dist = debug_camera[10]
            camera_target = np.array(debug_camera[11])

            camera_position = camera_target-dist*camera_forward
            ray_start_pos = camera_position
            far_plane = 10000
            ray_forward = camera_target-camera_position
            ray_forward = far_plane*ray_forward / np.linalg.norm(ray_forward)

            horizontal_distance = horizontal/width
            vertical_distance = vertical/height

            ray_end_pos = ray_start_pos + ray_forward - 0.5*horizontal + \
                0.5*vertical+float(mouse_x) * horizontal_distance - float(mouse_y)*vertical_distance
            ray_info = p.rayTest(ray_start_pos, ray_end_pos)
            hit = ray_info[0]
            return hit[0], hit[1]
    return -1, -1



import numpy as np
import pybullet as p

from pybullet_industrial.utility import draw_path, draw_coordinate_system


class ToolPath:
    """A Base object for representing a  toolpaths

    Args:
        positions (np.array(3,n)): A 3 dimensional array whith each dimension containing
                                   subsequent positions.
        orientations (np.array(4,n), optional): A 4 dimensional array where each dimension
                                                describes a subsequent quaternion.
                                                Defaults to None in which case
                                                the orientation [0,0,0,1] is assumed.
        tool_acivations (np.array(n), optional): A 1 dimensional array with boolean values
                                                 describing wheter a tool is active
                                                 at a given path pose.
                                                 Defaults to None in which case
                                                 the tool is always inactive.

    Raises:
        ValueError: If all given input arrays are different lengths.
    """

    def __init__(self, positions: np.array, orientations: np.array = None,
                 tool_acivations: np.array = None):

        self.positions = positions
        if orientations is None:
            self.orientations = np.zeros((4, len(self.positions[0])))
            self.orientations[3] = 1
        else:
            if len(orientations[0]) != len(positions[0]):
                raise ValueError(
                    "The position and orientation paths need to have the same length")
            self.orientations = orientations

        if tool_acivations is None:
            self.tool_activations = np.zeros(len(self.positions[0]))
        else:
            if len(tool_acivations[0]) != len(positions[0]):
                raise ValueError(
                    "The position and tool activation paths need to have the same length")
            self.tool_activations = tool_acivations

    def translate(self, vector: np.array):
        """Translates the whole path by a given vector

        Args:
            vector (np.array): A 3D vector describing the path translation
        """
        self.positions[0] += vector[0]
        self.positions[1] += vector[1]
        self.positions[2] += vector[2]

    def get_start_pose(self):
        """Returns the start pose of the trajectory for initial positioning

        Returns:
            np.array: a 3D position vector
            np.array: a 4D quaternion describing the orientation
        """
        return self.positions[:, 0], self.orientations[:, 0]

    def rotate(self, quaternion: np.array):
        """Rotates the vector by a given quaternion.
           Can be combined with pybullet.getQuaternionFromEuler() for easier usage.

        Args:
            quaternion (np.array): A 4 dimensional quaternion as a list or numpy array
        """
        path_positions = np.transpose(self.positions)
        path_orientations = np.transpose(self.orientations)

        rot_matrix = p.getMatrixFromQuaternion(quaternion)
        rot_matrix = np.array(rot_matrix).reshape(3, 3)
        for i in range(len(self)):
            path_positions[i] = rot_matrix@path_positions[i]
            _, path_orientations[i] = p.multiplyTransforms([0, 0, 0],
                                                           path_orientations[i],
                                                           [0, 0, 0],
                                                           quaternion)

        self.positions = np.transpose(path_positions)
        self.orientations = np.transpose(path_orientations)

    def draw(self, pose: bool = False, color: list = [0, 0, 1]):
        """Function which draws the path into the Debugin GUI.
           The path can either be a line representing the positions
           or a series of coordinate systems representing the whole pose.

        Args:
            orientation (bool, optional): Flag which determins if only the path position is shown
                                          or the full pose. Defaults to False.
            color (list, optional): The color of the line used for position only drawing.
                                    Defaults to [0, 0, 1].
        """
        if pose is False:
            draw_path(self.positions, color)
        else:
            path_positions = np.transpose(self.positions)
            path_orientations = np.transpose(self.orientations)
            for i in range(len(self)):
                draw_coordinate_system(
                    path_positions[i], path_orientations[i])

    def append(self, tool_path):
        """Appends a given ToolPath object to the end of this tool path.

        Args:
            tool_path (ToolPath): Another ToolPath object.
        """
        self.positions = np.append(self.positions, tool_path.positions, axis=1)
        self.orientations = np.append(
            self.orientations, tool_path.orientations, axis=1)
        self.tool_activations = np.append(
            self.tool_activations, tool_path.tool_activations)

    def prepend(self, tool_path):
        """Prepends a given ToolPath object to the start of this tool path.

        Args:
            tool_path (ToolPath): Another ToolPath object.
        """
        self.positions = np.append(tool_path.positions, self.positions, axis=1)
        self.orientations = np.append(
            tool_path.orientations, self.orientations, axis=1)
        self.tool_activations = np.append(
            tool_path.tool_activations, self.tool_activations)

    def __len__(self):
        return len(self.positions[0])

    def __iter__(self):
        self.current_index = 0
        return self

    def __next__(self):
        if self.current_index <= len(self)-1:
            i = self.current_index
            self.current_index += 1
            return self.positions[:, i], self.orientations[:, i], self.tool_activations[i]
        else:
            raise StopIteration



import numpy as np
import pybullet as p

from pybullet_industrial.endeffector_tool import EndeffectorTool


class Gripper(EndeffectorTool):
    """The base class for all Tools and Sensors connected to a Robot

    Args:
        urdf_model (str): A valid path to a urdf file describing the tool geometry
        start_position (np.array): the position at which the tool should be spawned
        start_orientation (np.array): the orientation at which the tool should be spawned
        coupled_robots (pi.RobotBase, optional): A wbk_sim.Robot object if
                                                 the robot is coupled from the start.
                                                 Defaults to None.
        tcp_frame (str, optional): The name of the urdf_link
                                        describing the tool center point.
                                        Defaults to None in which case the last link is used.
        connector_frames (str, optional): The name of the urdf_link
                                            at which a robot connects.
                                            Defaults to None in which case the base link is used.
    """

    def __init__(self, urdf_model: str, start_position, start_orientation,
                 coupled_robots=None, tcp_frame=None, connector_frames=None):

        super().__init__(urdf_model, start_position, start_orientation,
                         coupled_robots, tcp_frame, connector_frames)

        self._joint_name_to_index = {}
        self._lower_joint_limit = np.zeros(p.getNumJoints(self.urdf))
        self._upper_joint_limit = np.zeros(p.getNumJoints(self.urdf))
        self._actuated_joints = []

        for joint_number in range(p.getNumJoints(self.urdf)):
            if p.getJointInfo(self.urdf, joint_number)[2] != 4:
                joint_name = p.getJointInfo(self.urdf, joint_number)[
                    1].decode("utf-8")
                self._joint_name_to_index[joint_name] = joint_number
                self._actuated_joints.append(joint_number)

                lower_limit = p.getJointInfo(self.urdf, joint_number)[8]
                upper_limit = p.getJointInfo(self.urdf, joint_number)[9]
                if upper_limit < lower_limit:
                    lower_limit = -np.inf
                    upper_limit = np.inf
                self._lower_joint_limit[joint_number] = lower_limit
                self._upper_joint_limit[joint_number] = upper_limit

        self.max_joint_force = 20 * np.ones(p.getNumJoints(self.urdf))

        for joint_number in range(p.getNumJoints(self.urdf)):
            p.resetJointState(self.urdf, joint_number, targetValue=0)

        self._gripper_constraints = []
        for i in range(1, len(self._actuated_joints)):
            self._gripper_constraints.append(p.createConstraint(self.urdf,
                                                                self._actuated_joints[0],
                                                                self.urdf,
                                                                self._actuated_joints[i],
                                                                jointType=p.JOINT_GEAR,
                                                                jointAxis=[
                                                                    1, 0, 0],
                                                                parentFramePosition=[
                                                                    0, 0, 0],
                                                                childFramePosition=[0, 0, 0]))

            p.changeConstraint(
                self._gripper_constraints[-1], gearRatio=-1, erp=0.1, maxForce=5000)
        for joint_number in self._actuated_joints:
            p.setJointMotorControl2(self.urdf, joint_number, p.VELOCITY_CONTROL, targetVelocity=0,
                                    force=self.max_joint_force[joint_number])

    def actuate(self, target: float):
        """Actuates the gripper.

        Args:
            target(float): Relative amount of maximum travel distance the gripper should move
         """
        for joint_number in self._actuated_joints:
            p.setJointMotorControl2(self.urdf, joint_number, p.POSITION_CONTROL,
                                    force=self.max_joint_force[joint_number],
                                    targetPosition=self._lower_joint_limit[joint_number] + (
                                        self._upper_joint_limit[joint_number]
                                        - self._lower_joint_limit[joint_number]) * target)


class SuctionGripper(EndeffectorTool):
    """The base class for all Tools and Sensors connected to a Robot

    Args:
        urdf_model (str): A valid path to a urdf file describing the tool geometry
        start_position (np.array): the position at which the tool should be spawned
        start_orientation (np.array): the orientation at which the tool should be spawned
        coupled_robots (pi.RobotBase, optional): A wbk_sim.Robot object if
                                                 the robot is coupled from the start.
                                                 Defaults to None.
        tcp_frame (str, optional): The name of the urdf_link
                                   describing the tool center point.
                                   Defaults to None in which case the last link is used.
        connector_frames (str, optional): The name of the urdf_link
                                          at which a robot connects.
                                          Defaults to None in which case the base link is used.
        suction_links (str, optional): The names of the urdf_links wich represent
                                       the active suction parts.
                                       Defaults to all Links.

    """

    def __init__(self, urdf_model: str, start_position, start_orientation,
                 coupled_robots=None, tcp_frame=None, connector_frames=None, suction_links=None):

        super().__init__(urdf_model, start_position, start_orientation,
                         coupled_robots, tcp_frame, connector_frames)
        self.suction_constraints = []
        self._suction_links_ids = []

        self._link_name_to_index = {}
        for joint_number in range(p.getNumJoints(self.urdf)):
            link_name = p.getJointInfo(self.urdf, joint_number)[
                12].decode("utf-8")
            self._link_name_to_index[link_name] = joint_number

        if suction_links is None:
            self._suction_links_ids = None
        else:
            for name in suction_links:
                self._suction_links_ids.append(self._link_name_to_index[name])

    def activate(self, tolerance=0.0001):
        """Function to activate the suction gripper creating constraints between gripper and object

        Args:
            tolerance (float, optional): tolerance of contacts, i.e.
                                         at which distance the contact point is considered relevant.
        Returns:
            list[int]: ids of coupled objects
        """
        contact_points = list(p.getContactPoints(self.urdf))
        position_g, orientation_g = self.get_tool_pose()
        inv_pos, inv_orn = p.invertTransform(position_g, orientation_g)

        coupled_bodys = []

        for point in contact_points:
            if point[2] is self.urdf or point[2] is self._coupled_robot.urdf or \
                    point[8] > tolerance:
                del point

        if self._suction_links_ids is not None:
            for point in contact_points:
                if point[3] not in self._suction_links_ids:
                    del point

        for point in contact_points:
            if point[2] not in coupled_bodys:
                coupled_bodys.append(point[2])

        for body in coupled_bodys:
            position_o, orientation_o = p.getBasePositionAndOrientation(body)
            restraint_pos, restraint_orn = p.multiplyTransforms(
                inv_pos, inv_orn, position_o, orientation_o)
            cid_grip = p.createConstraint(self.urdf, self._tcp_id,
                                          body, -1,
                                          p.JOINT_FIXED, [0, 0, 0],
                                          parentFramePosition=restraint_pos,
                                          childFramePosition=[0, 0, 0],
                                          parentFrameOrientation=restraint_orn,
                                          childFrameOrientation=None)
            self.suction_constraints.append(cid_grip)
        return coupled_bodys

    def deactivate(self):
        """Function to deactivate the suction gripper
           deleting constraints between gripper and object
        """
        for constraint in self.suction_constraints:
            p.removeConstraint(constraint)



import numpy as np
import pybullet as p

from pybullet_industrial import RobotBase


class EndeffectorTool:
    """The base class for all Tools and Sensors connected to a Robot

    Args:
        urdf_model (str): A valid path to a urdf file describint the tool geometry
        start_position (np.array): the position at which the tool should be spawned
        start_orientation (np.array): the orientation at which the tool should be spawned
        coupled_robot (pi.RobotBase, optional): A pybullet_omdistrial.RobotBase object if
                                                the robot is coupled from the start.
                                                Defaults to None.
        tcp_frame (str, optional): The name of the urdf_link
                                   describing the tool center point.
                                   Defaults to None in which case the last link is used.
        connector_frame (str, optional): The name of the urdf_link at which a robot connects.
                                         Defaults to None in which case
                                         the base link is used.
    """

    def __init__(self, urdf_model: str, start_position: np.array, start_orientation: np.array,
                 coupled_robot: RobotBase = None, tcp_frame: str = None,
                 connector_frame: str = None):

        urdf_flags = p.URDF_USE_SELF_COLLISION_EXCLUDE_ALL_PARENTS
        self.urdf = p.loadURDF(urdf_model,
                               start_position, start_orientation,
                               flags=urdf_flags,
                               useFixedBase=False)

        self._link_name_to_index = {}
        self._coupled_robots = {}
        for joint_number in range(p.getNumJoints(self.urdf)):
            link_name = p.getJointInfo(self.urdf, joint_number)[
                12].decode("utf-8")
            self._link_name_to_index[link_name] = joint_number

        if tcp_frame is None:
            last_link = max(self._link_name_to_index)
            self._tcp_id = self._link_name_to_index[last_link]
        else:
            self._tcp_id = self._convert_link_to_id(tcp_frame)

        if connector_frame is None:
            self._connector_id = -1
            base_pos, base_ori = p.getBasePositionAndOrientation(self.urdf)
        else:
            self._connector_id = self._convert_link_to_id(connector_frame)
            link_state = p.getLinkState(self.urdf, self._connector_id)
            base_pos = link_state[0]
            base_ori = link_state[1]

        self._tcp_translation, self._tcp_rotation = p.multiplyTransforms(
            *p.invertTransform(base_pos, base_ori), *self.get_tool_pose(tcp_frame))

        self._coupled_robot = None
        self._coupling_link = None

        self._coupling_constraint = p.createConstraint(self.urdf,
                                                       -1, -1, -1,
                                                       p.JOINT_FIXED,
                                                       [0, 0, 0],
                                                       [0, 0, 0],
                                                       start_position,
                                                       None,
                                                       start_orientation)

        if coupled_robot is not None:
            self.couple(coupled_robot)

    def couple(self, robot: RobotBase, endeffector_name: str = None):
        """Dynamically Couples the Tool with the Endeffector of a given robot.
        Note that this endeffector can also be a virtual link to connect a sensor.
        A Tool can only be coupled with one robot

        Args:
            robot (pybullet_industrial.RobotBase): The robot whith which the tool should couple.
            endeffector_name (str, optional): The endeffector of the robot
                                              where the tool should couple to.
                                              Defaults to None.

        Raises:
            ValueError: If the tool is already coupled.
            TypeError: if the object to couple is nof of class RobotBase.
        """
        if self._coupled_robot is not None:
            raise ValueError("The tool is already coupled with a robot")
        if not isinstance(robot, RobotBase):
            raise TypeError(
                "A EndeffectorTool can only couple with a RobotBase object")

        if endeffector_name is None:
            endeffector_index = robot._default_endeffector_id
        else:

            endeffector_index = robot._convert_endeffector(
                endeffector_name)
        self._coupled_robot = robot
        self._coupling_link = endeffector_name
        p.removeConstraint(self._coupling_constraint)
        self._coupling_constraint = p.createConstraint(self._coupled_robot.urdf, endeffector_index,
                                                       self.urdf, self._connector_id,
                                                       p.JOINT_FIXED,
                                                       [0, 0, 0],
                                                       [0, 0, 0],
                                                       [0, 0, 0])

    def is_coupled(self):
        """Function which returns true if the Tool is currently coupled to a robot

        Returns:
            bool: 1 if the tool is coupled, 0 if not
        """
        if self._coupled_robot is None:
            return 0
        else:
            return 1

    def decouple(self):
        """Decouples the tool from the current robot.
           In this case a new constraint is created rooting the tool in its current pose.
        """
        self._coupled_robot = None
        self._coupling_link = None
        p.removeConstraint(self._coupling_constraint)
        position, orientation = p.getBasePositionAndOrientation(self.urdf)
        self._coupling_constraint = p.createConstraint(self.urdf,
                                                       -1, -1, -1,
                                                       p.JOINT_FIXED,
                                                       [0, 0, 0],
                                                       [0, 0, 0],
                                                       position,
                                                       None,
                                                       orientation)

    def get_tool_pose(self, tcp_frame: str = None):
        """Returns the pose of the tool center point.
           Using the tcp_frame argument the state of other links can also be returned

        Args:
            tcp_frame (str, optional): the name of the link whose pose should be returned.
                                       Defaults to None in which case the default tcp is used

        Returns:
            np.array: The 3D position the link the world coordinate system
            np.array: A quaternion describing the orientation of the link in world coordinates
        """
        if tcp_frame is None:
            tcp_id = self._tcp_id
        else:
            tcp_id = self._convert_link_to_id(tcp_frame)

        link_state = p.getLinkState(self.urdf, tcp_id)

        position = np.array(link_state[0])
        orientation = np.array(link_state[1])
        return position, orientation

    def set_tool_pose(self, target_position: np.array, target_orientation: np.array = None):
        """Allows the control of the tool.
           If the tool is coupled the inverse kinematic control of a coupled robot is used.
           If not the tool is moved directly.

        Args:
            target_position (np.array): the desired position of the tool center point (tcp)
            target_orientation (np.array, optional): the desired position of
                                                   the tool center point (tcp).
                                                   If none is provided only
                                                   the position of the robot is controlled.
        """

        if self.is_coupled():
            tcp_translation_inv, tcp_rotation_inv = p.invertTransform(
                self._tcp_translation, self._tcp_rotation)
            adj_target_position, adj_target_orientation = p.multiplyTransforms(
                target_position, target_orientation, tcp_translation_inv, tcp_rotation_inv)

            self._coupled_robot.set_endeffector_pose(
                adj_target_position, adj_target_orientation, endeffector_name=self._coupling_link)
        else:
            if target_orientation is None:
                _, adj_target_orientation = p.getBasePositionAndOrientation(
                    self.urdf)
            p.removeConstraint(self._coupling_constraint)
            self._coupling_constraint = p.createConstraint(self.urdf,
                                                           self._tcp_id, -1, -1,
                                                           p.JOINT_FIXED,
                                                           [0, 0, 0],
                                                           [0, 0, 0],
                                                           target_position,
                                                           None,
                                                           target_orientation)

    def apply_tcp_force(self, force: np.array, world_coordinates: bool = True):
        """Function which can apply a external Force at a the next simulation step.

            Carefull, this does not behave as expected for setRealTimeSimulation(1)!

        Args:
            force (np.array): A 3 dimensional force vector in Newton.
            world_coordinates (bool, optional): Specify wheter the force is defined
                                                in the world coordinates or the relative link frame.
                                                Defaults to True.
        """
        if world_coordinates:
            position, _ = self.get_tool_pose()
            p.applyExternalForce(self.urdf, self._tcp_id,
                                 force, position, p.WORLD_FRAME)
        else:
            p.applyExternalForce(self.urdf, self._tcp_id,
                                 force, [0, 0, 0], p.LINK_FRAME)

    def apply_tcp_torque(self, torque: np.array):
        """Function which can apply a external Torque at a the next simulation step.
           The local tcp_link frames are used as the main torque axis.

            Carefull, this does not behave as expected for setRealTimeSimulation(1)!

        Args:
            torque (np.array): A 3 dimensional torque vector in Newtonmeter.
        """
        p.applyExternalTorque(self.urdf, self._tcp_id,
                              torque, p.LINK_FRAME)

    def _convert_link_to_id(self, tcp: str):
        """Internal function that converts between link names and pybullet specific indexes

        Args:
            tcp (str): the name of the tool center point link

        Raises:
            TypeError: If the provided object is not a string

        Returns:
            int: the pybullet specific index of the link
        """
        if not isinstance(tcp, str):
            raise TypeError(
                "The Link name must be a String describing a URDF link")
        if not tcp in self._link_name_to_index:
            raise ValueError("Invalid Link name! valid names are: " +
                             str(self._link_name_to_index.keys()))

        return self._link_name_to_index[tcp]



# setup.py

import setuptools


setuptools.setup(
    name='pybullet_industrial',
    version='1.0',
    author='Jan Baumgärtner, Malte Hansjosten, Dominik Schönhofen',
    description='Pybullet_industrial is a process-aware robot simulation framework for pybullet.',
    long_description='Pybullet_industrial is a process-aware robot simulation framework for pybullet.',
    url='https://pybullet-industrial.readthedocs.io/en/latest/',
    license='MIT',
    install_requires=['numpy', 'pybullet', 'scipy'],
    packages=setuptools.find_packages(),

)
